/*
  mex interface to the iterative matrix library (IML++).  IML++ is a
  collection of iterative algorithms for solving linear systems of
  equations.  These methods are particularly well-suited for solving
  sparse systems.  Matlab includes most (maybe all) of these
  algorithms, but they are implemented as m-files, so they are quite
  slow.  See http://math.nist.gov/iml++/ for details and
  documentation.  IML++ was developed by the National Institute of
  Standards and Technology, and is in the public domain.
*/
#include "mex.h" /* header file for matlab API */
/* IML++ header files */
#include "iml/bicg.h"
#include "iml/bicgstab.h"
#include "iml/cg.h"
#include "iml/cgs.h"
#include "iml/ir.h"
#include "iml/qmr.h"

#include <string.h>

class Vector {  
  double *val;
  int siz;
  bool mx;
public: 
  Vector() {val=NULL; siz=0; mx=true;} 
  Vector(unsigned int n) { 
    val= (double*) mxCalloc(n, sizeof(double));
    siz = n;
    mx = false;
  }
  Vector(const mxArray *marr) {
    val = mxGetPr(marr); 
    siz = mxGetN(marr)*mxGetM(marr);
    mx = true;
  }
  Vector(const Vector &x) {
    siz = x.size();
    mx = false;
    val= (double*) mxCalloc(siz, sizeof(double));
    memcpy(val,x.val,sizeof(double)*siz);
  }
  ~Vector() {
    if (!mx) {
      mxFree(val);    
      siz = 0;
      val = NULL;
    }
  }
  int size() const { return siz; }
  Vector& operator=(const double &x) { 
    for(int i=0;i<siz;i++) val[i] = x;
    return *this;
  } 
  Vector& operator=(const Vector &x) {
    if (siz!=x.size()) {
      if (!mx) mxFree(val);      
      siz = x.size();
      mx = false;
      val = (double*) mxCalloc(siz, sizeof(double));
    }
    memcpy(val,x.val,sizeof(double)*siz);
    return *this;
  }
  Vector operator+(const Vector &x) const { 
    Vector y(siz); 
    for(unsigned int i=0;i<siz;i++) y[i] = val[i]+x[i];
    return y;
  }
  Vector& operator+=(const Vector &x){ 
    for(unsigned int i=0;i<siz;i++) val[i] += x[i];
    return *this;
  }
  Vector operator-(const Vector &x) const { 
    Vector y(siz);
    for(unsigned int i=0;i<siz;i++) y[i] = val[i] - x[i];
    return y;
  }
  Vector& operator-=(const Vector &x){ 
    for(unsigned int i=0;i<siz;i++) val[i] -= x[i];
    return *this;
  }

  Vector operator*(const double &a) const {
    Vector y(this->size());
    for(unsigned int i=0;i<siz;i++) y[i] = val[i]*a;
    return y;
  }
  friend Vector operator*(const double &a,const Vector &x) {
    Vector y(x.size());
    for(unsigned int i=0;i<y.size();i++) y[i] = x[i]*a;
    return y;    
  }
  double& operator()(int i) { 
    if (i>=siz || i<0) mexErrMsgTxt("badIndex");
    return val[i]; 
  }
  double& operator()(unsigned int i) {
    if (i>=siz || i<0) mexErrMsgTxt("badIndex");
    return val[i]; 
  }
  double& operator[](int i) { 
    if (i>=siz || i<0) mexErrMsgTxt("badIndex");
    return val[i]; 
  }
  const double& operator[](int i) const { 
    if (i>=siz || i<0) mexErrMsgTxt("badIndex");
    return val[i]; 
  }
  double& operator[](unsigned int i) { 
    if (i>=siz || i<0) mexErrMsgTxt("badIndex");
    return val[i]; 
  }
  friend double dot(const Vector &x,const Vector &y) {
    double dp = 0;
    for(unsigned int i=0;i<y.size();i++) dp += y[i]*x[i];
    return dp;
  }
  friend double norm(const Vector &x) {
    return(sqrt(dot(x,x)));
  }
}; 

class SparseMatrix
{
public:
  double *val; /* values */
  int *ir,*jc; /* indices */
  int nnz;
  int ncol;
  SparseMatrix() { };
  SparseMatrix(const mxArray *marr) {    
    if (!mxIsSparse(marr)) {
      mexErrMsgTxt("Sparse matrix required.");
    }
    nnz = mxGetNzmax(marr);
    ir = mxGetIr(marr);
    jc = mxGetJc(marr);
    ncol = mxGetN(marr);
    val = mxGetPr(marr);
  }
  Vector operator*(const Vector &x) const {
    Vector y(x.size());
    int i = 0;
    y = 0.0;   
    for(int c=0;c<ncol;c++) {
      while(i < jc[c+1]) {
        y[ir[i]] += val[i]*x[c]; 
        i++;
      }
    }
    return y;
  }
  Vector trans_mult(const Vector &x) const {
    Vector y(x.size());
    int i=jc[0];
    y = 0.0;
    for(int c=0;c<ncol;c++) {
      while(i < jc[c+1]) {
        y[c] += val[i]*x[ir[i]]; i++;
      }
    }
    return y;
  }
};

class Precond {
public:
  Precond() { }
  Vector solve(const Vector &x) const {
    Vector y(x);
    return(y);
  }
  Vector trans_solve(const Vector &x) const {
    Vector y(x);
    return(y);
  }
};


void mexFunction
(int nlhs, /* number of left hand side arguments */
 mxArray *plhs[], /* pointer to lhs*/
 int nrhs, /* number of rhs arguments*/
 const mxArray *prhs[]) /* pointer to rhs arguments*/
{
  /* rhs args: A, b, tol, maxit, method */
  int maxit = 100; /* default max iterations */
  double tol = 1e-6; /* default tolerance */
  char *method=NULL;
  mxArray *xmx = mxCreateDoubleMatrix(mxGetN(prhs[0]),1,mxREAL);
  SparseMatrix A(prhs[0]);
  Vector b(prhs[1]);
  Vector x(xmx);
  x = 0;
  int exitFlag;
  Precond M;
  if (nrhs>=3) tol = mxGetScalar(prhs[2]);
  if (nrhs>=4) maxit = (int) mxGetScalar(prhs[3]);
  if (nrhs>=5) {
    int nchar = mxGetN(prhs[4])*mxGetM(prhs[4])+2;
    method = (char*) mxCalloc(nchar, sizeof(char));
    mxGetString(prhs[4],method,nchar);
  }
  if (method==NULL || strcmp(method,"bicg")==0) {
    exitFlag = BiCG(A,x,b,M,maxit,tol);
  } else if (strcmp(method,"bicgstab")==0) {
    exitFlag = BiCGSTAB(A,x,b,M,maxit,tol);
  } else if (strcmp(method,"cg")==0) {
    exitFlag = CG(A,x,b,M,maxit,tol);
  } else if (strcmp(method,"cgs")==0) {
    exitFlag = CGS(A,x,b,M,maxit,tol);
  } else if (strcmp(method,"ir")==0) {
    exitFlag = IR(A,x,b,M,maxit,tol);
  } else if (strcmp(method,"qmr")==0) {
    exitFlag = QMR(A,x,b,M,M,maxit,tol);
  } else {
    mexErrMsgTxt("method unrecognized");
  }
  if (nlhs>0) plhs[0] = xmx;
  if (nlhs>1) plhs[1] = mxCreateDoubleScalar(exitFlag);
}
