#include <math.h>
#include "mex.h"
#define EPSILON 1.110326e-16

double Sample_Double_Truncated_Normal(const double mu,const double var,const double a,const double b);

void mexFunction
( 
 int nlhs, /* number of left hand side arguments */
 mxArray *plhs[], /* pointer to lhs*/
 int nrhs, /* number of rhs arguments*/
 const mxArray *prhs[]) /* pointer to rhs arguments*/
{
  /* input pointers */
  /* all inputs should be vectors of the same size */
  /* returns a vector of draws from the correspaonding distributions */
  double *mu;    /* mean */
  double *var;   /* variance */
  double *lo;    /* lower bound */
  double *hi;    /* upper bound */
  
  /* output */
  double *x;     /* random draws */
  
  int N; /* number of draws */
  int i,m,n; /* loop indices */
  /* check number of input arguments */
  if (nrhs!=4) mexErrMsgTxt("Need 4 inputs");   
  /* assign input pointers */
  mu = mxGetPr(prhs[0]);
  var = mxGetPr(prhs[1]);
  lo = mxGetPr(prhs[2]);
  hi = mxGetPr(prhs[3]);
  /*N = (*mxGetPr(prhs[4]));*/
  /* mexErrMsgTxt(""); */
  /* check that inputs all same size */
  m = mxGetM(prhs[0]);
  n = mxGetN(prhs[0]);
  N = m>n? m:n;
  for(i=1;i<4;i++) {
    m = mxGetM(prhs[i]);
    n = mxGetN(prhs[i]);
    if (N != (m>n? m:n)) {
      mexErrMsgTxt("Input arguments must all be same size");
    }
    if (!(m==1 || n==1)) {
      mexErrMsgTxt("Input arguments should be vectors");
    }
  }
  
  /* allocate the output vector */
  plhs[0] = mxCreateDoubleMatrix(N,1,mxREAL);
  x = mxGetPr(plhs[0]);
  /* draw the numbers */
  for(i=0;i<N;i++) {
    x[i] = Sample_Double_Truncated_Normal(*mu,*var,*lo,*hi);
    /*mexPrintf("i = %d, x = %g\n",i,x[i]);*/
  }
  return;
}

/*/////////////////////////////////////////////// RANDOM NUMBER GENERATION //////////////////////////////////////////////*/
/*  Parameters for: 
    Marsaglia & Tsang generator for random normals & random exponentials. 
    Marsaglia, G. & Tsang, W.W. (2000) `The ziggurat method for generating random variables', J. Statist. Software, 
    v5(8). This is an electronic journal which can be downloaded from:  http://www.jstatsoft.org/v05/i08 */
static unsigned int jz,jsr=123456789;
static int seed_initialized=0;
#define SHR3 (jz=jsr, jsr^=(jsr<<13), jsr^=(jsr>>17), jsr^=(jsr<<5),jz+jsr)
#define UNI (.5 + (signed) SHR3*.2328306e-9)
#define IUNI SHR3
static int hz;
static unsigned int iz, kn[128], ke[256];
static double wn[128],fn[128], we[256],fe[256];
#define RNOR (hz=SHR3, iz=hz&127, (fabs(hz)<kn[iz])? hz*wn[iz] : nfix())
#define REXP (jz=SHR3, iz=jz&255, (jz <ke[iz])? jz*we[iz] : efix())

double nfix(void)
{ /* nfix() generates variates from the residue when rejection in RNOR occurs. */
  const double r = 3.442620f;     /* The start of the right tail */
  static double x, y;
  for(;;) {
    x=hz*wn[iz];      /* iz==0, handles the base strip */
    if(iz==0) {
      do {
	x=-log(UNI)*0.2904764; /* .2904764 is 1/r */
	y=-log(UNI);
      } while(y+y<x*x);
      return (hz>0)? r+x : -r-x;
    }
    /* iz>0, handle the wedges of other strips */
    if ( fn[iz]+UNI*(fn[iz-1]-fn[iz]) < exp(-.5*x*x) ) return x;
    /* initiate, try to exit for(;;) for loop*/
    hz=SHR3;
    iz=hz&127;
    if(fabs(hz)<kn[iz]) return (hz*wn[iz]);
  }
}

double efix(void)
{ /* efix() generates variates from the residue when rejection in REXP occurs.  */
  double x;
  for(;;) {
    if(iz==0) return (7.69711-log(UNI));          /* iz==0 */
    x=jz*we[iz];
    if( fe[iz]+UNI*(fe[iz-1]-fe[iz]) < exp(-x) ) return (x);
    /* initiate, try to exit for(;;) loop */
    jz=SHR3;
    iz=(jz&255);
    if(jz<ke[iz]) return (jz*we[iz]);
  }
}

unsigned int get_jsr(void)
{
  return(jsr);
}

void set_jsr(unsigned int jsrin)
{
  jsr = jsrin;
  return;
}

/* This helps when you want to start the seuqence at the same place again */
void reset_seed(unsigned int jsrseed)
{ /* This procedure sets the seed and creates the tables.   */
  /************************** THIS IS WHAT MAKES IT DIFFERENT FROM SET_SEED ****************************************/
  jsr = 123456789; 
  /*****************************************************************************************************************/
  jsr^=jsrseed;
}

void set_seed(unsigned int jsrseed)
{ /* This procedure sets the seed and creates the tables.  */
  const double m1 = 2147483648.0, m2 = 4294967296.;
  double dn=3.442619855899,tn=dn,vn=9.91256303526217e-3, q;
  double de=7.697117470131487, te=de, ve=3.949659822581572e-3;
  int i;
  jsr^=jsrseed;
  /* Set up tables for RNOR */
  q=vn/exp(-.5*dn*dn);
  kn[0]=(int)((dn/q)*m1);
  kn[1]=0;
  wn[0]=q/m1;
  wn[127]=dn/m1;
  fn[0]=1.;
  fn[127]=exp(-.5*dn*dn);
  for(i=126;i>=1;i--) {
    dn=sqrt(-2.*log(vn/dn+exp(-.5*dn*dn)));
    kn[i+1]=(int)((dn/tn)*m1);
    tn=dn;
    fn[i]=exp(-.5*dn*dn);
    wn[i]=dn/m1;
  }
  /* Set up tables for REXP */
  q = ve/exp(-de);
  ke[0]=(int)((de/q)*m2);
  ke[1]=0;
  we[0]=q/m2;
  we[255]=de/m2;
  fe[0]=1.;
  fe[255]=exp(-de);
  for(i=254;i>=1;i--) {
    de=-log(ve/de+exp(-de));
    ke[i+1]= (int)((de/te)*m2);
    te=de;
    fe[i]=exp(-de);
    we[i]=de/m2;
  }
  seed_initialized=1;
}

#define T1 0.375
#define T2 2.18
#define T3 0.725
#define T4 0.45
#define F(x) exp(-.5*(x*x))
double Sample_Double_Truncated_Normal(const double mu,const double var,const double a,const double b)
{ /* Generates one draw from truncated standard normal distribution (mu,sigma) on (a,b) */
  double c,c1,c2,u[2],x,cdel,f1,f2,z,az,bz;
  int lflip,j;
  unsigned int seed;
  mxArray *plhs[1], *prhs[1];
  if (seed_initialized==0) {
    mexCallMATLAB(1,plhs,0,prhs,"rand");
    seed = mxGetScalar(plhs[0]);
    set_seed(seed);
  }
  c1=az=(a-mu)/sqrt(var);
  c2=bz=(b-mu)/sqrt(var);
  lflip=0;
  if (c1*c2<0.0) {
    if ((F(c1)>T1) && (F(c2)>T1)) {
      cdel=c2-c1;
      do {
	for(j=0;j<=1;j++) u[j] = UNI;
	x=c1+cdel*u[0];
      } while (!(u[1]<F(x)));
    }
    else
      do {
	x=RNOR;
      } while (!((x>c1) && (x<c2)));
  }
  else {
    if (c1<0.0) {
      c=c1;
      c1=-c2;
      lflip=1;
    }
    f1=F(c1);
    f2=F(c2);
    if ( (f2<EPSILON) || (f1/f2>T2)) {
      if (c1>T3){
	/* exponential rejection sampling */
	c=c2-c1;
	do {
	  u[0] = UNI;
	  z = REXP/c1;
	} while (!((z<c) && (u[0]<F(z))));
	x=c1+z;
      }
      else {
	/* half-normal rejection sampling */
	do {
	  x=RNOR;
	  x=fabs(x);
	} while (!((x>c1) && (x<c2)));
      }
    }
    else {
      /* uniform rejection sampling */
      cdel=c2-c1;
      do {
	for(j = 0;j<= 1;j++) u[j] = UNI;
	x=c1+cdel*u[0];
      } while (!(u[1] < F(x)/f1));
    }
  }
  return lflip ? (mu-sqrt(var)*x):(mu+sqrt(var)*x);
}
#undef F
#undef T1
#undef T2
#undef T3
#undef T4
